import logging
import adage
import adage.node
import adage.serialize
import adage.nodestate
import jsonpointer
import time
import os
import datetime
import jsonpath_rw
import yadagestep

from backends import NoneProxy
from helpers import get_obj_id, get_id_fromjson
from yadagestep import outputReference

from handlers.predicate_handlers import handlers as pred_handlers

log = logging.getLogger(__name__)

class stage_base(object):
    '''
    Base class for workflow stages (i.e. extension rules)
    provides common datastructures and the extension predicate.
    Implementations are required to provide a schedule()
    method that is called upon apply
    '''

    def __init__(self, name, context, dependencies=None):
        self.view = None
        self.name = name
        self.context = context
        self.depspec = dependencies

    def applicable(self, flowview):
        if not self.depspec:
            return True
        predicate = pred_handlers[self.depspec['dependency_type']]
        return predicate(flowview, self.depspec)

    def apply(self, flowview):
        self.view = flowview
        self.schedule()

    def schedule(self):
        raise NotImplementedError()

    def addStep(self, step):
        dependencies = [self.view.dag.getNode(k.stepid) for k in step.inputs]

        try:
            depwrites = []
            for d in dependencies:
                try:
                    depwrites += d.task.context['readwrite']
                except AttributeError:
                    pass
            step.context['depwrites'] = depwrites
        except AttributeError:
            pass
        return self.view.addStep(step, stage = self.name, depends_on=dependencies)

    def addWorkflow(self, rules, initstep):
        self.view.addWorkflow(rules, initstep=initstep, stage=self.name)

    #(de-)serialization
    def json(self):
        return {
            'name': self.name,
            'context': self.context,
            'dependencies': self.depspec
        }


class initStage(stage_base):
    '''
    simple stage that just adds a initializer step to the DAG
    '''

    def __init__(self, step, context, dependencies):
        super(initStage, self).__init__('init', context, dependencies)
        self.step = step

    def schedule(self):
        log.debug('initializing a scope with init step: %s',
                  self.step.prepublished)
        self.addStep(self.step)

    #(de-)serialization
    @classmethod
    def fromJSON(cls, data):
        instance = cls(
            step=yadagestep.initstep.fromJSON(data['step']),
            context=data['context'],
            dependencies=data['dependencies']
        )
        return instance

    def json(self):
        data = super(initStage, self).json()
        data.update(type='initStage', info='', step=self.step.json())
        return data


class jsonStage(stage_base):
    '''
    A stage that is defined via the JSON scheduler schemas
    '''

    def __init__(self, json, context):
        self.stageinfo = json['scheduler']
        super(jsonStage, self).__init__(
            json['name'], context, json['dependencies'])

    def __repr__(self):
        return '<jsonStage: {}>'.format(self.name)

    def schedule(self):
        from handlers.scheduler_handlers import handlers as sched_handlers
        scheduler = sched_handlers[self.stageinfo['scheduler_type']]
        scheduler(self, self.stageinfo)

    #(de-)serialization
    @classmethod
    def fromJSON(cls, data):
        return cls(json={
            'scheduler': data['info'],
            'name': data['name'],
            'dependencies': data['dependencies']
        }, context=data['context'])

    def json(self):
        data = super(jsonStage, self).json()
        data.update(type='jsonStage', info=self.stageinfo)
        return data


class offsetRule(object):
    '''
    A wrapper object around a scoped rule, so that it can be applied from
    a global p.o.v., i.e. as adage expects its rules.
    '''

    def __init__(self, rule, offset=None, identifier=None):
        '''
        initializes a scoped rule. scope is defined by a JSONPointer, e.g.
        one[0]two.three

        '''

        self.rule = rule
        self.offset = offset
        self.identifier = identifier or get_id_fromjson({
            'rule': rule.json(),
            'offset': offset
        })

    def __repr__(self):
        return '<offsetStage {}/{} >'.format(self.offset,self.rule.name)

    def applicable(self, adageobj):
        '''
        determin whether the rule is applicable. Evaluated within the offset.
        :param adageobj: the workflow object
        '''
        x = self.rule.applicable(WorkflowView(adageobj, self.offset))
        return x

    def apply(self, adageobj):
        '''
        applies a rule within the scope set by offset

        :param adageobj: the workflow object
        '''
        self.rule.apply(WorkflowView(adageobj, self.offset))

    #(de-)serialization
    @classmethod
    def fromJSON(cls, data):
        if data['rule']['type'] == 'initStage':
            rule = initStage.fromJSON(data['rule'])
        elif data['rule']['type'] == 'jsonStage':
            rule = jsonStage.fromJSON(data['rule'])
        return cls(
            rule=rule,
            identifier=data['id'],
            offset=data['offset']
        )

    def json(self):
        return {'type': 'offset',
                'id': self.identifier,
                'offset': self.offset,
                'rule': self.rule.json()}


class YadageNode(adage.node.Node):
    '''
    Node object for yadage that extends the default with
    the ability to have prepublished results
    '''

    def __init__(self, name, task, identifier=None):
        super(YadageNode, self).__init__(name, task, identifier)

    def __repr__(self):
        lifetime = datetime.timedelta(seconds = (time.time() - self.define_time))
        runtime = None
        if self.state != adage.nodestate.DEFINED:
            referencetime = time.time() if not self.ready() else self.ready_by_time
            runtime = datetime.timedelta(seconds = (referencetime - self.submit_time))
        return '<YadageNode {} {} lifetime: {}  runtime: {} (id: {}) has result: {}>'.format(
            self.name, self.state, lifetime, runtime, self.identifier, self.has_result()
        )

    def has_result(self):
        if 'YADAGE_IGNORE_PREPUBLISHING' in os.environ:
            return self.successful()
        return (self.task.prepublished is not None) or self.successful()

    @property
    def result(self):
        if self.task.prepublished is not None and 'YADAGE_IGNORE_PREPUBLISHING' not in os.environ:
            if self.ready() and self.successful():
                sanity =  super(YadageNode, self).result == self.task.prepublished
                if not sanity:
                    raise RuntimeError('prepublished and actual result differ:\nlhs:\n{}\nrhs:{}'.format(
                        super(YadageNode, self).result,self.task.prepublished)
                )
            return self.task.prepublished
        return super(YadageNode, self).result

    def readfromresult(self,pointerpath, whoisreading = None, failsilently = False):
        if not self.has_result():
            if failsilently: return None
            raise RuntimeError('attempt')
        pointer = jsonpointer.JsonPointer(pointerpath)
        if whoisreading:
            whoisreading.inputs.append(outputReference(self.identifier,pointer))
        v = pointer.resolve(self.result)
        return v

    @classmethod
    def fromJSON(cls, data):
        if data['task']['type'] == 'initstep':
            task = yadagestep.initstep.fromJSON(data['task'])
        elif data['task']['type'] == 'yadagestep':
            task = yadagestep.yadagestep.fromJSON(data['task'])
        return cls(data['name'], task, data['id'])


def json_or_nil(x):
    return None if x is None else x.json()


class YadageWorkflow(adage.adageobject):
    '''
    The overall workflow state object that extends the basic
    Adage state object by two bookkeeping structures.
    '''

    def __init__(self):
        super(YadageWorkflow, self).__init__()
        self.stepsbystage = {}
        self.bookkeeping = {}

    def view(self, offset=''):
        return WorkflowView(self, offset)

    def json(self):
        from adage.serialize import obj_to_json
        data = obj_to_json(self,
                           ruleserializer=json_or_nil,
                           taskserializer=json_or_nil,
                           proxyserializer=json_or_nil,
                           )
        data['bookkeeping'] = self.bookkeeping
        data['stepsbystage'] = self.stepsbystage
        return data

    #(de-)serialization
    @classmethod
    def fromJSON(cls, data, proxyclass=NoneProxy, backend=None):
        rules, applied = [], []
        for x in data['rules']:
            rules += [offsetRule.fromJSON(x)]
        for x in data['applied']:
            applied += [offsetRule.fromJSON(x)]

        instance = cls()
        instance.rules = rules
        instance.applied_rules = applied
        instance.bookkeeping = data['bookkeeping']
        instance.stepsbystage = data['stepsbystage']
        instance.dag = adage.serialize.dag_from_json(
            data['dag'],
            YadageNode,
            proxyclass,
            backend
        )
        return instance

    @classmethod
    def createFromJSON(cls, jsondata, context):
        instance = cls()
        # rules = [offsetRule(jsonStage(yml,context), offset = '') for yml in jsondata['stages']]
        rules = [jsonStage(yml, context) for yml in jsondata['stages']]
        instance.view().addWorkflow(rules)
        return instance


def createOffsetMeta(offset, bookkeeping):
    '''
    sets up a location to track rule and step ids for a given scope offset
    '''
    pointer = jsonpointer.JsonPointer(offset)
    view = bookkeeping
    for x in pointer.parts:
        if x not in view:
            view[x] = {}
        view = view[x]
    scoped = pointer.resolve(bookkeeping)
    if '_meta' not in scoped:
        scoped['_meta'] = {'rules': [], 'steps': []}


class WorkflowView(object):
    '''
    Provides a 'view' of the overall workflow object that corresponds
    to a particular level of nesting. That is, this presents the scope
    within which extension rules operate (i.e. add steps, reference
    other steps, etc)
    '''

    def __init__(self, workflowobj, offset=''):
        self.dag = workflowobj.dag
        self.rules = workflowobj.rules
        self.applied_rules = workflowobj.applied_rules
        self.offset = offset
        self.steps = jsonpointer.JsonPointer(
            self.offset).resolve(workflowobj.stepsbystage)
        self.bookkeeper = jsonpointer.JsonPointer(
            self.offset).resolve(workflowobj.bookkeeping)

    def query(self, query, collection):
        matches = jsonpath_rw.parse(query).find(collection)
        return matches

    def getSteps(self, query):
        '''
        returns steps related to the JSONPath query.
        if a query points to a stage (say 'stagename'):
            will return all toplevel steps.. no recursion into subworkflows
        if a query points to steps (e.g. 'stagename[*]')
            will return a steps directly

        '''
        nodeids = []

        matches = self.query(query, self.steps)
        for match in matches:
            value = match.value
            #step endpoint case
            if isinstance(value,dict) and '_nodeid' in value: 
                nodeids.append(value['_nodeid'])
            #stage endpoint case
            elif isinstance(value,list):
                for item in value:
                    if '_nodeid' in item:
                        nodeids.append(item['_nodeid'])

        result = [self.dag.getNode(nodeid) for nodeid in nodeids]
        return result

    def _makeoffset(self, offset):
        '''
        prepare a full offset based on this views' offset and a relative offset

        :param offset: the relative offset
        '''
        thisoffset = jsonpointer.JsonPointer(offset)
        if self.offset:
            fulloffset = jsonpointer.JsonPointer.from_parts(
                jsonpointer.JsonPointer(self.offset).parts + thisoffset.parts).path
        else:
            fulloffset = thisoffset.path
        return fulloffset

    def getRule(self, name=None, offset='', identifier=None):
        '''retrieve a rule by offset or name or identifier'''
        fulloffset = self._makeoffset(offset)
        for x in self.rules + self.applied_rules:
            if x.identifier == identifier or (x.offset == fulloffset and x.rule.name == name):
                return x
        return None

    def init(self, initdata, name='init'):
        '''
        initialize this scope by adding an initialization stage.

        :param inidata: initialization JSON data
        '''
        step = yadagestep.initstep(name, initdata)
        self.addRule(initStage(step, {}, None), self.offset)

    def addRule(self, rule, offset=''):
        '''
        add a DAG extensloaderion rule, possibly with a scope offset
        '''
        thisoffset = jsonpointer.JsonPointer(offset)
        offsetrule = offsetRule(rule, self._makeoffset(offset))
        self.rules += [offsetrule]
        createOffsetMeta(thisoffset.path, self.bookkeeper)
        thisoffset.resolve(self.bookkeeper)['_meta'][
            'rules'] += [offsetrule.identifier]
        return offsetrule.identifier

    def addStep(self, step, stage, depends_on=None):
        '''
        adds a node to the DAG connecting it to the passed depending nodes
        while tracking that it was added by the specified stage
        '''
        node = YadageNode(step.name, step, identifier=get_obj_id(step))
        self.dag.addNode(node, depends_on=depends_on)

        log.debug('added node %s', node)

        noderef = {'_nodeid': node.identifier}
        self.steps.setdefault(stage,[]).append(noderef)
        self.bookkeeper['_meta']['steps'] += [node.identifier]
        return node

    def addWorkflow(self, rules, initstep=None, stage=None):
        if initstep:
            rules += [initStage(initstep, {}, None)]
        newsteps = {}
        if stage in self.steps:
            self.steps[stage] += [newsteps]
        elif stage is not None:
            self.steps[stage] = [newsteps]

        offset = jsonpointer.JsonPointer.from_parts(
            [stage, len(self.steps[stage]) - 1]).path if stage else ''
        if stage is not None:
            self.steps[stage][-1]['_offset'] = offset

        for rule in rules:
            self.addRule(rule, offset)



#!/usr/bin/env python
import click
import os
import steering_api
import logging
import yadageschemas
import yaml
import clihelpers

log = logging.getLogger(__name__)


@click.command()
@click.option('-v', '--verbosity', default='INFO', help = 'logging verbosity')
@click.option('-i', '--loginterval', default=30, help = 'adage tracking interval in seconds')
@click.option('-u', '--updateinterval', default=0.02, help = 'adage graph inspection interval in seconds')
@click.option('-m', '--schemasource', default=yadageschemas.schemadir, help = 'schema directory for workflow validation')
@click.option('-b', '--backend', default='multiproc:auto', help = 'packtivity backend string')
@click.option('-c', '--cache', default='')
@click.option('-s', '--statectrl', default='inmem')
@click.option('-d','--initdir', default='init', help = "relative path (to workdir) to initialiation data directory")
@click.option('--interactive/--not-interactive', default=False, help = 'en-/disable user interactio (sign-off graph extensions and packtivity submissions)')
@click.option('--validate/--no-validate', default=True, help = 'en-/disable workflow spec validation')
@click.option('--accept-workdir/--no-accept-workdir', default=False)
# v below this we should only have options/arg available also in yadage-manual
@click.option('--visualize/--no-visualize', default=True, help = 'visualize workflow graph')
@click.option('-t', '--toplevel', default=os.getcwd(), help = 'toplevel uri to be used to resolve workflow name and references from')
@click.option('-r', '--read', default=None, help = 'YAML file to initialize the state context')
@click.option('-a', '--inputarchive', default=None, help = 'initial data to stage as input')
@click.option('--parameter', '-p', multiple=True, help = '<parameter name>=<yaml string> input parameter specifcations ')
@click.argument('workdir')
@click.argument('workflow')
@click.argument('initfiles', nargs=-1)
def main(workdir,
         workflow,
         initfiles,
         toplevel,
         verbosity,
         loginterval,
         updateinterval,
         schemasource,
         backend,
         interactive,
         parameter,
         validate,
         read,
         visualize,
         inputarchive,
         statectrl,
         cache,
         accept_workdir,
         initdir):

    logging.basicConfig(level=getattr(logging, verbosity), format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    if inputarchive:
        initdir = clihelpers.prepare_workdir_from_archive(workdir, inputarchive)
    else:
        initdir = os.path.join(workdir,initdir)

    initdata = clihelpers.getinit_data(initfiles, parameter)
    backend  = clihelpers.setupbackend_fromstring(backend, cacheconfig=cache)

    rc = steering_api.run_workflow(
        workdir,
        workflow,
        initdata,
        toplevel,
        backend = backend,
        initdir = initdir,
        updateinterval = updateinterval,
        loginterval = loginterval,
        read = yaml.load(open(read)) if read else None,
        validate = validate,
        doviz = visualize,
        schemadir = schemasource,
        user_interaction = interactive,
        accept_existing_workdir = accept_workdir,
        ctrlsetup = statectrl
    )
    if rc:
        exc = click.exceptions.ClickException(
            click.style("Workflow failed", fg='red'))
        exc.exit_code = rc
        raise exc

if __name__ == '__main__':
    main()



import hashlib
import uuid
import json
import os
import jsonref
import copy
from backends.trivialbackend import TrivialProxy, TrivialBackend

import logging
import jsonpointer

log = logging.getLogger(__name__)

def leaf_iterator(jsonlike, path = None):
    ''''
    :param jsonlike: a jsonlike object (i.e. nested lists/arrays: leaf-types must not be JSONable)
    :return: iterator that yields (JSON path, leaf value) tuples
    '''
    path = path or []
    if type(jsonlike) == list:
        for i,x in enumerate(jsonlike):
            thispath = copy.deepcopy(path)
            thispath.append(i)
            for leaf in leaf_iterator(x, path = thispath):
                yield leaf
    elif type(jsonlike) == dict:
        for k,v in jsonlike.iteritems():
            thispath = copy.deepcopy(path)
            thispath.append(k)
            for leaf in leaf_iterator(v, path = thispath):
                yield leaf
    else:
        yield jsonpointer.JsonPointer.from_parts(path),jsonlike


def set_backend(dag, backend, proxymaker):
    '''
    sets backend and proxies for each node in the DAG.
    proxymaker is a 1-ary function that takes the node object and
    returns a suitable result proxy
    '''
    for nodename in dag.nodes():
        n = dag.getNode(nodename)
        n.backend = backend
        n.resultproxy = proxymaker(n)


def set_trivial_backend(dag, jsondata):
    '''
    sets the backend for the case of a static backend
    Proxies are set using the node identifier
    '''
    backend = TrivialBackend()
    set_backend(
        dag,
        backend,
        proxymaker=lambda n: TrivialProxy(
            status=jsondata[n.identifier]['status'],
            result=jsondata[n.identifier]['result']
        )
    )
    return backend


def stepindex(wflow):
    ''''
    build an index mapping step id's to rule id's
    '''
    index = {}
    for r in wflow.applied_rules + wflow.rules:
        for s in wflow.view(r.offset).getSteps(r.rule.name):
            index[s.identifier] = r.identifier
    return index

DEFAULT_ID_METHOD = 'jsonhash'


def json_hash(jsonable):
    # log.info('hashing: %s',json.dumps(jsonable, cls=WithJsonRefEncoder, sort_keys=True))
    the_hash  = hashlib.sha1(json.dumps(jsonable, cls=WithJsonRefEncoder, sort_keys=True)).hexdigest()
    # log.info('got %s',hash)
    return the_hash


def get_id_fromjson(jsonobject, method = DEFAULT_ID_METHOD):
    method = os.environ.get('YADAGE_ID_METHOD', method)
    if method == 'uuid':
        return str(uuid.uuid1())
    elif method == 'jsonhash':
        return json_hash(jsonobject)
    else:
        raise NotImplementedError(
            'unkown id generation method {}'.format(method))

def get_obj_id(obj_with_json_method, method = DEFAULT_ID_METHOD):
    return get_id_fromjson(obj_with_json_method.json(), method)


class WithJsonRefEncoder(json.JSONEncoder):

    def default(self, obj):
        if isinstance(obj, jsonref.JsonRef):
            return {k: v for k, v in obj.iteritems()}
        else:
            super(WithJsonRefEncoder, self).default(obj)



import yadageschemas


def workflow(source, toplevel, schema_name='yadage/workflow-schema', schemadir=None, validate=True):
    data = yadageschemas.load(source, toplevel, schema_name, schemadir, validate)
    return data



import federatedbackend
import logging
import yadage.yadagestep
from trivialbackend import TrivialProxy, TrivialBackend
from packtivity.backendutils import backend_from_string

log = logging.getLogger(__name__)


class InitProxy(TrivialProxy):
    def proxyname(self):
        return 'InitProxy'


class PacktivityBackend(federatedbackend.FederatedBackend):
    '''
    a backend that mainly submits step tasks to packtivity backend
    except for init nodes, which are resolvable trivially.
    '''

    def __init__(self,  packtivity_backendstring = None, packtivity_backend = None, cacheconfig = None):
        if packtivity_backendstring:
            is_sync, backend = backend_from_string(packtivity_backendstring)
            assert not is_sync
        elif packtivity_backend:
            backend = packtivity_backend
        else:
            raise RuntimeError('need backend or backendstring')
        if cacheconfig:
            self.cached = True
            import caching
            backend = caching.CachedBackend(
                backend,
                cacheconfig=cacheconfig
            )
        else:
            self.cached = False
        super(PacktivityBackend, self).__init__({
            'init': TrivialBackend(),
            'packtivity': backend
        })

    def routedsubmit(self, task):
        tasktype = type(task)
        if tasktype == yadage.yadagestep.yadagestep:
            #this is a little hacky, because the packtivity backends
            #take unrolled spec/parameters/context while the adage API
            #takes generalized task objects
            #possibly could use Munch on the packtivity side to
            #dynammicaly create .task/.attributes/.context-able objects
            if self.cached:
                #Cached backends adhere to the task-based API
                return self.backends['packtivity'].submit(task)
            else:
                #primary packtivity backends adhere to the unrolled API
                return self.backends['packtivity'].submit(
                    task.spec, task.attributes, task.context
                )
        elif tasktype == yadage.yadagestep.initstep:
            #init steps are by definition successful
            return InitProxy(status = 'SUCCESS', result = task.attributes)

    def routeproxy(self, proxy):
        if type(proxy) == InitProxy:
            return 'init'
        else:
            return 'packtivity'
        raise NotImplementedError('needs implementation')



#!/usr/bin/env python

from contextlib import contextmanager
from steering_object import YadageSteering
import logging
import importlib
import interactive
import yadageschemas
import os
log = logging.getLogger(__name__)

RC_FAILED = 1
RC_SUCCEEDED = 0


def run_workflow(*args, **kwargs):
    """
    Main entry point to run a Yadage workflow
    """
    # let's be conservative and just assume we're going to fail. will set
    # success RC explicityly
    return_value = RC_FAILED
    try:
        with steering_ctx(*args, **kwargs):
            pass
        return_value = RC_SUCCEEDED
    except:
        log.exception('Unfortunately we failed. :(')
    return return_value

@contextmanager
def steering_ctx(
    workdir,
    workflow,
    initdata,
    loadtoplevel,
    backend,
    read = None,
    initdir = None,
    updateinterval = 0.02,
    loginterval = 30,
    schemadir = yadageschemas.schemadir,
    user_interaction=False,
    validate=True,
    doviz=True,
    accept_existing_workdir = False,
    ctrlsetup = 'inmem'):
    
    ys = YadageSteering()
    ys.prepare_workdir(workdir, accept_existing_workdir, contextinit = read)
    ys.init_workflow(workflow, loadtoplevel, initdata, ctrlsetup = ctrlsetup, initdir = initdir, validate = validate, schemadir = schemadir)
    
    custom_tracker = os.environ.get('YADAGE_CUSTOM_TRACKER',None)
    if custom_tracker:
        modulename,trackerclassname = custom_tracker.split(':')
        module = importlib.import_module(modulename)
        trackerclass = getattr(module,trackerclassname)
        ys.adage_argument(additional_trackers = [trackerclass()])

    ys.adage_argument(
        default_trackers = doviz,
        trackevery = loginterval,
        update_interval = updateinterval,
        workdir='{}/_adage'.format(ys.workdir)
    )
    if user_interaction:
        extend, submit = interactive.interactive_deciders()
        ys.adage_argument(
            extend_decider = extend,
            submit_decider = submit
        )
    yield ys

    log.info('running yadage workflow %s on backend %s', workflow, backend)
    try:
        ys.run_adage(backend)
    finally:
        ys.serialize()
    if doviz:
        ys.visualize()

